/*  Copyright (C) 2008 Chris Hellyar <chris@trash.co.nz>
 *
 *  This software is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *                   
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *                                  
 *  You should have received a copy of the GNU General Public License
 *  along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307  USA.  (See COPYING file) */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

// shared libraries I'm using...
#include <modbus.h>
#include <confuse.h>

#include <mysql/mysql.h>

// stuff for this only.
#include "readconf.c"

// I'm not using this anywhere, but it might come in handy!
#define VERSION = 0.50

// globals (Yuck)
modbus_param_t mb_param; // structure for modbus stuff.
MYSQL mysql;

// prototypes.
void DoLog(void);	// routine that does the logging
int CheckLogFile(void); // checks the log file access before starting

int main(int argc,char *argv[])
{
	int pid,Forked;

	if(ReadConfig()) {  // read the config in from the config file setup is in readconf.c
		printf("Error Reading Config - Exiting\n");
		exit(EXIT_FAILURE);
	}
        if(Verbose) printf("Successfully read Config file\n");

	// open the mysql database if required...
	if(strlen(cfg_getstr(cfg,"DBServer"))!=0) {
		if(Verbose) printf("Config had mysql options - opening database\n");
		mysql_init(&mysql);
		if (!mysql_real_connect(&mysql,
			cfg_getstr(cfg,"DBServer"),
			cfg_getstr(cfg,"DBUser"),
			cfg_getstr(cfg,"DBPassword"),
			cfg_getstr(cfg,"DBName"),
			3306,NULL,0))
		{
			fprintf(stderr, "Failed to connect to database: Error: %s\n",
			mysql_error(&mysql));
			exit(EXIT_FAILURE);
		}
	} else if(Verbose) printf("No DBServer=\"\" no mysql connection required\n");

	// check if the log file is writable, returns null if not...
	if(!CheckLogFile()) exit(EXIT_FAILURE);

	if(Verbose) printf("!-----  Verbose Mode enabled Lots of output will follow ------!\n");

	printf("Using Serial Modbus on %s\n",cfg_getstr(cfg,"Port"));

        /* RTU parity : none, even, odd */
        modbus_init_rtu(&mb_param, cfg_getstr(cfg,"Port"), 19200, "even", 8, 1);
        modbus_set_debug(&mb_param, Verbose);  // set debug messages on/off from config file Verbose setting.
        if(modbus_connect(&mb_param)) {
		printf("Failed to initialise modbus interface (serial port)\n");
		exit(EXIT_FAILURE);
	} else if(Verbose) printf("Success initialising modbus interface (serial port)\n");

	// make it a deamon, sorta.
	if(cfg_getbool(cfg,"Daemon")) {
		pid = fork();	// fork the thing already.

		if ( pid == -1 )    {
			printf("\n\n Failed to fork().\nProgram will now exit.\n");
			exit( EXIT_FAILURE );
		}
		if ( pid == 0 ) { // We are the child, so detach ourselves.
			setpgid( 0,0 );
			printf("Program has become a daemon.\n");
			close( STDIN_FILENO );      // We don't have a shell to write to now
			close( STDOUT_FILENO );     // so these should be closed.
			close( STDERR_FILENO );
		} else { // We're the Parent so exit.
			sleep(1);
			exit( EXIT_SUCCESS );
		}
	}

	if(cfg_getint(cfg,"Interval")!=0) {
		while(1) {
			DoLog();
			sleep(cfg_getint(cfg,"Interval"));
		}
	} else DoLog();

	return 0;
}

/* DoLog - reads in from the modbus RTU, and writes to 
 * log file if it's configured. */
void DoLog(void) {  // log once, like.
	int i;
	FILE * fp;
	char Output[256]; // string for outputting CSV (If enabled)
	char Buffer[256]; // Buffer String
	char LogFile[256]; // Logfile name....
	int Inputs,Address,DeviceID,Interval;
	char Name[256];   // Name
	int InTemp[1];    // temp array for reading inputs
	float Results[32];  // the results go here, like.

	Inputs = cfg_size(cfg, "input"); // Number of inputs to scan
	Interval = cfg_getint(cfg, "Interval"); // Number of inputs to scan
	strcpy(LogFile,cfg_getstr(cfg, "LogFile")); // Logfile name

	Output[0]=0;  // pre-trim Output string.

	if(!Daemon & Verbose) printf("Scanning %d Inputs\n",Inputs);

	for(i = 0; i < Inputs; i++)  // do the loopy loop.
	{
		cfg_t *arg = cfg_getnsec(cfg, "input", i);  // reads next 'input' block in the config file.

		// get the stuff out of the config for this input.
		strcpy(Name,cfg_getstr(arg,"Name"));
		Address = cfg_getint(arg,"Address");
		DeviceID = cfg_getint(arg,"DeviceID");

		if(!Daemon & Verbose) printf("Reading input \"%s\" Address [%d] from modbus Device [%d]\n",
			            Name, Address, DeviceID);

		// decide if it's discrete or register the go read it...
		if(Address>9999 & Address <20000) { // Discrete
			if(!Daemon & Verbose) printf("Reading single discrete input\n");
			read_input_status(&mb_param, DeviceID, Address-10001, 1, InTemp);
			if(!Daemon & Verbose) printf("Got value [%d]\n",InTemp[0]);
			Results[i]=InTemp[0];  // stuff it in the 'output' array.
		} else if(Address>29999 & Address < 40000) { // Register
			if(!Daemon & Verbose) printf("Reading single input Register\n");
			read_input_registers(&mb_param, DeviceID, Address-30001, 1, InTemp);
			if(!Daemon & Verbose) printf("Got value [%d]\n",InTemp[0]);
			// put it in the results, with scaling and offset.
			Results[i] = InTemp[0]*cfg_getfloat(arg,"Scale")+cfg_getfloat(arg,"Offset"); 
		} else {
			printf("Address invalid\n");
			exit(EXIT_FAILURE);
		}
	}

	// Create output string.
	for(i = 0; i < Inputs; i++) {
		sprintf(Buffer,"%.4f,",Results[i]);
		strcat(Output,Buffer);
	}

	// create the time stamp string thingit to go on the end of the output
	time_t rawtime;
	time ( &rawtime );
	if(cfg_getbool(cfg,"UnixTime")) sprintf(Buffer,"%ld\n",rawtime);  // stick in unix or ctime() timestamp.
	else sprintf(Buffer,"%ld",rawtime);  // stick in unix or ctime() timestamp.
	strcat(Output,Buffer);
	if(!Daemon) printf("Output: %s",Output);

	// write mysql entry if we're doing that.
	if(strlen(cfg_getstr(cfg,"DBServer"))!=0) {  // if we're using mysql...
		char Query[100];  // string for query passed tomysql
		for(i = 0; i < Inputs; i++) {
			sprintf(Query,"insert into log set inputid=%d,time=%d,value=%.4f",i,rawtime,Results[i]);
			if(!Daemon & Verbose) printf("mysql Query: %s\n",Query);
			mysql_query(&mysql,Query);
		}
	}

	if(strlen(LogFile)!=0 && Interval!=0) {  // If passed a logfile name and Interval, log the results with timestamp.
		if(!Daemon & Verbose) printf ("Writing to log file\n");
		fp=fopen(LogFile,"a");
		if(fp==NULL) { // open the file and give an error if it shites itself.
			if(!Daemon) printf("Failed to open logfile [%s]\nExiting",LogFile);
			exit(EXIT_FAILURE);
		}
		fputs(Output,fp);
		fclose(fp);
	}
}


/* CheckLogFile -----------------------------
 * Attempts to open the log file if one is specified
 * returns FALSE if it fails, and shows a message.
 * returns TRUE if no Interval && logfile specified 
 * or the the test open of the file works. */
int CheckLogFile(void) {
	int Interval=cfg_getint(cfg,"Interval");  // get the interval
	char LogFile[200];
	strcpy(LogFile,cfg_getstr(cfg,"LogFile"));  // get the LogFile name

	if(strlen(LogFile)!=0 && Interval!=0) {  // If passed a logfile name and Interval, log the results with timestamp.
		if(Verbose) printf("Interval=%d LogFile=\"%s\" - Checking log file.\n",Interval,LogFile);
		FILE * fp=fopen(LogFile,"a");
		if(fp==NULL) { // open the file and give an error if it shites itself.
			printf("Failed to open logfile [%s] (Check Permissions etc)\nExiting\n",LogFile);
			return(FALSE); // About here it's borked...
		} else {
			if(Verbose) printf("LogFileOpened OK\n");
			fclose(fp);  // if it opened, close it.
		}
	} else {
		if (Verbose) printf("LogFile or Interval not set - Not checking log file\n");
	}
	return(TRUE);  // everything is OK if e get here.
}	

